@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import socket
 import ssl
 import sys
@@ -5,12 +7,18 @@ from ipaddress import IPv6Address, ip_address
 from os import PathLike, chmod
 from pathlib import Path
 from socket import AddressFamily, SocketKind
-from typing import Awaitable, List, Optional, Tuple, Union, cast, overload
+from typing import Awaitable, List, Tuple, cast, overload
 
 from .. import to_thread
 from ..abc import (
-    ConnectedUDPSocket, IPAddressType, IPSockAddrType, SocketListener, SocketStream, UDPSocket,
-    UNIXSocketStream)
+    ConnectedUDPSocket,
+    IPAddressType,
+    IPSockAddrType,
+    SocketListener,
+    SocketStream,
+    UDPSocket,
+    UNIXSocketStream,
+)
 from ..streams.stapled import MultiListener
 from ..streams.tls import TLSStream
 from ._eventloop import get_asynclib
@@ -23,20 +31,28 @@ if sys.version_info >= (3, 8):
 else:
     from typing_extensions import Literal
 
-IPPROTO_IPV6 = getattr(socket, 'IPPROTO_IPV6', 41)  # https://bugs.python.org/issue29515
+IPPROTO_IPV6 = getattr(socket, "IPPROTO_IPV6", 41)  # https://bugs.python.org/issue29515
 
-GetAddrInfoReturnType = List[Tuple[AddressFamily, SocketKind, int, str, Tuple[str, int]]]
-AnyIPAddressFamily = Literal[AddressFamily.AF_UNSPEC, AddressFamily.AF_INET,
-                             AddressFamily.AF_INET6]
+GetAddrInfoReturnType = List[
+    Tuple[AddressFamily, SocketKind, int, str, Tuple[str, int]]
+]
+AnyIPAddressFamily = Literal[
+    AddressFamily.AF_UNSPEC, AddressFamily.AF_INET, AddressFamily.AF_INET6
+]
 IPAddressFamily = Literal[AddressFamily.AF_INET, AddressFamily.AF_INET6]
 
 
 # tls_hostname given
 @overload
 async def connect_tcp(
-    remote_host: IPAddressType, remote_port: int, *, local_host: Optional[IPAddressType] = ...,
-    ssl_context: Optional[ssl.SSLContext] = ..., tls_standard_compatible: bool = ...,
-    tls_hostname: str, happy_eyeballs_delay: float = ...
+    remote_host: IPAddressType,
+    remote_port: int,
+    *,
+    local_host: IPAddressType | None = ...,
+    ssl_context: ssl.SSLContext | None = ...,
+    tls_standard_compatible: bool = ...,
+    tls_hostname: str,
+    happy_eyeballs_delay: float = ...,
 ) -> TLSStream:
     ...
 
@@ -44,9 +60,14 @@ async def connect_tcp(
 # ssl_context given
 @overload
 async def connect_tcp(
-    remote_host: IPAddressType, remote_port: int, *, local_host: Optional[IPAddressType] = ...,
-    ssl_context: ssl.SSLContext, tls_standard_compatible: bool = ...,
-    tls_hostname: Optional[str] = ..., happy_eyeballs_delay: float = ...
+    remote_host: IPAddressType,
+    remote_port: int,
+    *,
+    local_host: IPAddressType | None = ...,
+    ssl_context: ssl.SSLContext,
+    tls_standard_compatible: bool = ...,
+    tls_hostname: str | None = ...,
+    happy_eyeballs_delay: float = ...,
 ) -> TLSStream:
     ...
 
@@ -54,10 +75,15 @@ async def connect_tcp(
 # tls=True
 @overload
 async def connect_tcp(
-    remote_host: IPAddressType, remote_port: int, *, local_host: Optional[IPAddressType] = ...,
-    tls: Literal[True], ssl_context: Optional[ssl.SSLContext] = ...,
-    tls_standard_compatible: bool = ..., tls_hostname: Optional[str] = ...,
-    happy_eyeballs_delay: float = ...
+    remote_host: IPAddressType,
+    remote_port: int,
+    *,
+    local_host: IPAddressType | None = ...,
+    tls: Literal[True],
+    ssl_context: ssl.SSLContext | None = ...,
+    tls_standard_compatible: bool = ...,
+    tls_hostname: str | None = ...,
+    happy_eyeballs_delay: float = ...,
 ) -> TLSStream:
     ...
 
@@ -65,10 +91,15 @@ async def connect_tcp(
 # tls=False
 @overload
 async def connect_tcp(
-    remote_host: IPAddressType, remote_port: int, *, local_host: Optional[IPAddressType] = ...,
-    tls: Literal[False], ssl_context: Optional[ssl.SSLContext] = ...,
-    tls_standard_compatible: bool = ..., tls_hostname: Optional[str] = ...,
-    happy_eyeballs_delay: float = ...
+    remote_host: IPAddressType,
+    remote_port: int,
+    *,
+    local_host: IPAddressType | None = ...,
+    tls: Literal[False],
+    ssl_context: ssl.SSLContext | None = ...,
+    tls_standard_compatible: bool = ...,
+    tls_hostname: str | None = ...,
+    happy_eyeballs_delay: float = ...,
 ) -> SocketStream:
     ...
 
@@ -76,26 +107,35 @@ async def connect_tcp(
 # No TLS arguments
 @overload
 async def connect_tcp(
-    remote_host: IPAddressType, remote_port: int, *, local_host: Optional[IPAddressType] = ...,
-    happy_eyeballs_delay: float = ...
+    remote_host: IPAddressType,
+    remote_port: int,
+    *,
+    local_host: IPAddressType | None = ...,
+    happy_eyeballs_delay: float = ...,
 ) -> SocketStream:
     ...
 
 
 async def connect_tcp(
-    remote_host: IPAddressType, remote_port: int, *, local_host: Optional[IPAddressType] = None,
-    tls: bool = False, ssl_context: Optional[ssl.SSLContext] = None,
-    tls_standard_compatible: bool = True, tls_hostname: Optional[str] = None,
-    happy_eyeballs_delay: float = 0.25
-) -> Union[SocketStream, TLSStream]:
+    remote_host: IPAddressType,
+    remote_port: int,
+    *,
+    local_host: IPAddressType | None = None,
+    tls: bool = False,
+    ssl_context: ssl.SSLContext | None = None,
+    tls_standard_compatible: bool = True,
+    tls_hostname: str | None = None,
+    happy_eyeballs_delay: float = 0.25,
+) -> SocketStream | TLSStream:
     """
     Connect to a host using the TCP protocol.
 
-    This function implements the stateless version of the Happy Eyeballs algorithm (RFC 6555).
-    If ``address`` is a host name that resolves to multiple IP addresses, each one is tried until
-    one connection attempt succeeds. If the first attempt does not connected within 250
-    milliseconds, a second attempt is started using the next address in the list, and so on.
-    On IPv6 enabled systems, an IPv6 address (if available) is tried first.
+    This function implements the stateless version of the Happy Eyeballs algorithm (RFC
+    6555). If ``remote_host`` is a host name that resolves to multiple IP addresses,
+    each one is tried until one connection attempt succeeds. If the first attempt does
+    not connected within 250 milliseconds, a second attempt is started using the next
+    address in the list, and so on. On IPv6 enabled systems, an IPv6 address (if
+    available) is tried first.
 
     When the connection has been established, a TLS handshake will be done if either
     ``ssl_context`` or ``tls_hostname`` is not ``None``, or if ``tls`` is ``True``.
@@ -119,7 +159,7 @@ async def connect_tcp(
 
     """
     # Placed here due to https://github.com/python/mypy/issues/7057
-    connected_stream: Optional[SocketStream] = None
+    connected_stream: SocketStream | None = None
 
     async def try_connect(remote_host: str, event: Event) -> None:
         nonlocal connected_stream
@@ -138,7 +178,7 @@ async def connect_tcp(
             event.set()
 
     asynclib = get_asynclib()
-    local_address: Optional[IPSockAddrType] = None
+    local_address: IPSockAddrType | None = None
     family = socket.AF_UNSPEC
     if local_host:
         gai_res = await getaddrinfo(str(local_host), None)
@@ -149,13 +189,14 @@ async def connect_tcp(
         addr_obj = ip_address(remote_host)
     except ValueError:
         # getaddrinfo() will raise an exception if name resolution fails
-        gai_res = await getaddrinfo(target_host, remote_port, family=family,
-                                    type=socket.SOCK_STREAM)
+        gai_res = await getaddrinfo(
+            target_host, remote_port, family=family, type=socket.SOCK_STREAM
+        )
 
         # Organize the list so that the first address is an IPv6 address (if available) and the
         # second one is an IPv4 addresses. The rest can be in whatever order.
         v6_found = v4_found = False
-        target_addrs: List[Tuple[socket.AddressFamily, str]] = []
+        target_addrs: list[tuple[socket.AddressFamily, str]] = []
         for af, *rest, sa in gai_res:
             if af == socket.AF_INET6 and not v6_found:
                 v6_found = True
@@ -171,7 +212,7 @@ async def connect_tcp(
         else:
             target_addrs = [(socket.AF_INET, addr_obj.compressed)]
 
-    oserrors: List[OSError] = []
+    oserrors: list[OSError] = []
     async with create_task_group() as tg:
         for i, (af, addr) in enumerate(target_addrs):
             event = Event()
@@ -181,14 +222,17 @@ async def connect_tcp(
 
     if connected_stream is None:
         cause = oserrors[0] if len(oserrors) == 1 else asynclib.ExceptionGroup(oserrors)
-        raise OSError('All connection attempts failed') from cause
+        raise OSError("All connection attempts failed") from cause
 
     if tls or tls_hostname or ssl_context:
         try:
-            return await TLSStream.wrap(connected_stream, server_side=False,
-                                        hostname=tls_hostname or str(remote_host),
-                                        ssl_context=ssl_context,
-                                        standard_compatible=tls_standard_compatible)
+            return await TLSStream.wrap(
+                connected_stream,
+                server_side=False,
+                hostname=tls_hostname or str(remote_host),
+                ssl_context=ssl_context,
+                standard_compatible=tls_standard_compatible,
+            )
         except BaseException:
             await aclose_forcefully(connected_stream)
             raise
@@ -196,7 +240,7 @@ async def connect_tcp(
     return connected_stream
 
 
-async def connect_unix(path: Union[str, PathLike]) -> UNIXSocketStream:
+async def connect_unix(path: str | PathLike[str]) -> UNIXSocketStream:
     """
     Connect to the given UNIX socket.
 
@@ -211,41 +255,55 @@ async def connect_unix(path: Union[str, PathLike]) -> UNIXSocketStream:
 
 
 async def create_tcp_listener(
-    *, local_host: Optional[IPAddressType] = None, local_port: int = 0,
-    family: AnyIPAddressFamily = socket.AddressFamily.AF_UNSPEC, backlog: int = 65536,
-    reuse_port: bool = False
+    *,
+    local_host: IPAddressType | None = None,
+    local_port: int = 0,
+    family: AnyIPAddressFamily = socket.AddressFamily.AF_UNSPEC,
+    backlog: int = 65536,
+    reuse_port: bool = False,
 ) -> MultiListener[SocketStream]:
     """
     Create a TCP socket listener.
 
     :param local_port: port number to listen on
-    :param local_host: IP address of the interface to listen on. If omitted, listen on all IPv4
-        and IPv6 interfaces. To listen on all interfaces on a specific address family, use
-        ``0.0.0.0`` for IPv4 or ``::`` for IPv6.
-    :param family: address family (used if ``interface`` was omitted)
-    :param backlog: maximum number of queued incoming connections (up to a maximum of 2**16, or
-        65536)
-    :param reuse_port: ``True`` to allow multiple sockets to bind to the same address/port
-        (not supported on Windows)
+    :param local_host: IP address of the interface to listen on. If omitted, listen on
+        all IPv4 and IPv6 interfaces. To listen on all interfaces on a specific address
+        family, use ``0.0.0.0`` for IPv4 or ``::`` for IPv6.
+    :param family: address family (used if ``local_host`` was omitted)
+    :param backlog: maximum number of queued incoming connections (up to a maximum of
+        2**16, or 65536)
+    :param reuse_port: ``True`` to allow multiple sockets to bind to the same
+        address/port (not supported on Windows)
     :return: a list of listener objects
 
     """
     asynclib = get_asynclib()
     backlog = min(backlog, 65536)
     local_host = str(local_host) if local_host is not None else None
-    gai_res = await getaddrinfo(local_host, local_port, family=family,  # type: ignore[arg-type]
-                                type=socket.SOCK_STREAM,
-                                flags=socket.AI_PASSIVE | socket.AI_ADDRCONFIG)
-    listeners: List[SocketListener] = []
+    gai_res = await getaddrinfo(
+        local_host,  # type: ignore[arg-type]
+        local_port,
+        family=family,
+        type=socket.SocketKind.SOCK_STREAM if sys.platform == "win32" else 0,
+        flags=socket.AI_PASSIVE | socket.AI_ADDRCONFIG,
+    )
+    listeners: list[SocketListener] = []
     try:
         # The set() is here to work around a glibc bug:
         # https://sourceware.org/bugzilla/show_bug.cgi?id=14969
-        for fam, *_, sockaddr in sorted(set(gai_res)):
+        sockaddr: tuple[str, int] | tuple[str, int, int, int]
+        for fam, kind, *_, sockaddr in sorted(set(gai_res)):
+            # Workaround for an uvloop bug where we don't get the correct scope ID for
+            # IPv6 link-local addresses when passing type=socket.SOCK_STREAM to
+            # getaddrinfo(): https://github.com/MagicStack/uvloop/issues/539
+            if sys.platform != "win32" and kind is not SocketKind.SOCK_STREAM:
+                continue
+
             raw_socket = socket.socket(fam)
             raw_socket.setblocking(False)
 
             # For Windows, enable exclusive address use. For others, enable address reuse.
-            if sys.platform == 'win32':
+            if sys.platform == "win32":
                 raw_socket.setsockopt(socket.SOL_SOCKET, socket.SO_EXCLUSIVEADDRUSE, 1)
             else:
                 raw_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
@@ -257,6 +315,11 @@ async def create_tcp_listener(
             if fam == socket.AF_INET6:
                 raw_socket.setsockopt(IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
 
+                # Workaround for #554
+                if "%" in sockaddr[0]:
+                    addr, scope_id = sockaddr[0].split("%", 1)
+                    sockaddr = (addr, sockaddr[1], 0, int(scope_id))
+
             raw_socket.bind(sockaddr)
             raw_socket.listen(backlog)
             listener = asynclib.TCPSocketListener(raw_socket)
@@ -271,8 +334,11 @@ async def create_tcp_listener(
 
 
 async def create_unix_listener(
-        path: Union[str, PathLike], *, mode: Optional[int] = None,
-        backlog: int = 65536) -> SocketListener:
+    path: str | PathLike[str],
+    *,
+    mode: int | None = None,
+    backlog: int = 65536,
+) -> SocketListener:
     """
     Create a UNIX socket listener.
 
@@ -309,14 +375,17 @@ async def create_unix_listener(
 
 
 async def create_udp_socket(
-    family: AnyIPAddressFamily = AddressFamily.AF_UNSPEC, *,
-    local_host: Optional[IPAddressType] = None, local_port: int = 0, reuse_port: bool = False
+    family: AnyIPAddressFamily = AddressFamily.AF_UNSPEC,
+    *,
+    local_host: IPAddressType | None = None,
+    local_port: int = 0,
+    reuse_port: bool = False,
 ) -> UDPSocket:
     """
     Create a UDP socket.
 
-    If ``port`` has been given, the socket will be bound to this port on the local machine,
-    making this socket suitable for providing UDP based services.
+    If ``local_port`` has been given, the socket will be bound to this port on the local
+    machine, making this socket suitable for providing UDP based services.
 
     :param family: address family (``AF_INET`` or ``AF_INET6``) – automatically determined from
         ``local_host`` if omitted
@@ -331,23 +400,33 @@ async def create_udp_socket(
         raise ValueError('Either "family" or "local_host" must be given')
 
     if local_host:
-        gai_res = await getaddrinfo(str(local_host), local_port, family=family,
-                                    type=socket.SOCK_DGRAM,
-                                    flags=socket.AI_PASSIVE | socket.AI_ADDRCONFIG)
+        gai_res = await getaddrinfo(
+            str(local_host),
+            local_port,
+            family=family,
+            type=socket.SOCK_DGRAM,
+            flags=socket.AI_PASSIVE | socket.AI_ADDRCONFIG,
+        )
         family = cast(AnyIPAddressFamily, gai_res[0][0])
         local_address = gai_res[0][-1]
     elif family is AddressFamily.AF_INET6:
-        local_address = ('::', 0)
+        local_address = ("::", 0)
     else:
-        local_address = ('0.0.0.0', 0)
+        local_address = ("0.0.0.0", 0)
 
-    return await get_asynclib().create_udp_socket(family, local_address, None, reuse_port)
+    return await get_asynclib().create_udp_socket(
+        family, local_address, None, reuse_port
+    )
 
 
 async def create_connected_udp_socket(
-    remote_host: IPAddressType, remote_port: int, *,
+    remote_host: IPAddressType,
+    remote_port: int,
+    *,
     family: AnyIPAddressFamily = AddressFamily.AF_UNSPEC,
-    local_host: Optional[IPAddressType] = None, local_port: int = 0, reuse_port: bool = False
+    local_host: IPAddressType | None = None,
+    local_port: int = 0,
+    reuse_port: bool = False,
 ) -> ConnectedUDPSocket:
     """
     Create a connected UDP socket.
@@ -368,24 +447,36 @@ async def create_connected_udp_socket(
     """
     local_address = None
     if local_host:
-        gai_res = await getaddrinfo(str(local_host), local_port, family=family,
-                                    type=socket.SOCK_DGRAM,
-                                    flags=socket.AI_PASSIVE | socket.AI_ADDRCONFIG)
+        gai_res = await getaddrinfo(
+            str(local_host),
+            local_port,
+            family=family,
+            type=socket.SOCK_DGRAM,
+            flags=socket.AI_PASSIVE | socket.AI_ADDRCONFIG,
+        )
         family = cast(AnyIPAddressFamily, gai_res[0][0])
         local_address = gai_res[0][-1]
 
-    gai_res = await getaddrinfo(str(remote_host), remote_port, family=family,
-                                type=socket.SOCK_DGRAM)
+    gai_res = await getaddrinfo(
+        str(remote_host), remote_port, family=family, type=socket.SOCK_DGRAM
+    )
     family = cast(AnyIPAddressFamily, gai_res[0][0])
     remote_address = gai_res[0][-1]
 
-    return await get_asynclib().create_udp_socket(family, local_address, remote_address,
-                                                  reuse_port)
+    return await get_asynclib().create_udp_socket(
+        family, local_address, remote_address, reuse_port
+    )
 
 
-async def getaddrinfo(host: Union[bytearray, bytes, str], port: Union[str, int, None], *,
-                      family: Union[int, AddressFamily] = 0, type: Union[int, SocketKind] = 0,
-                      proto: int = 0, flags: int = 0) -> GetAddrInfoReturnType:
+async def getaddrinfo(
+    host: bytearray | bytes | str,
+    port: str | int | None,
+    *,
+    family: int | AddressFamily = 0,
+    type: int | SocketKind = 0,
+    proto: int = 0,
+    flags: int = 0,
+) -> GetAddrInfoReturnType:
     """
     Look up a numeric IP address given a host name.
 
@@ -409,20 +500,24 @@ async def getaddrinfo(host: Union[bytearray, bytes, str], port: Union[str, int,
     # Handle unicode hostnames
     if isinstance(host, str):
         try:
-            encoded_host = host.encode('ascii')
+            encoded_host = host.encode("ascii")
         except UnicodeEncodeError:
             import idna
+
             encoded_host = idna.encode(host, uts46=True)
     else:
         encoded_host = host
 
-    gai_res = await get_asynclib().getaddrinfo(encoded_host, port, family=family, type=type,
-                                               proto=proto, flags=flags)
-    return [(family, type, proto, canonname, convert_ipv6_sockaddr(sockaddr))
-            for family, type, proto, canonname, sockaddr in gai_res]
+    gai_res = await get_asynclib().getaddrinfo(
+        encoded_host, port, family=family, type=type, proto=proto, flags=flags
+    )
+    return [
+        (family, type, proto, canonname, convert_ipv6_sockaddr(sockaddr))
+        for family, type, proto, canonname, sockaddr in gai_res
+    ]
 
 
-def getnameinfo(sockaddr: IPSockAddrType, flags: int = 0) -> Awaitable[Tuple[str, str]]:
+def getnameinfo(sockaddr: IPSockAddrType, flags: int = 0) -> Awaitable[tuple[str, str]]:
     """
     Look up the host name of an IP address.
 
@@ -480,9 +575,10 @@ def wait_socket_writable(sock: socket.socket) -> Awaitable[None]:
 # Private API
 #
 
+
 def convert_ipv6_sockaddr(
-    sockaddr: Union[Tuple[str, int, int, int], Tuple[str, int]]
-) -> Tuple[str, int]:
+    sockaddr: tuple[str, int, int, int] | tuple[str, int]
+) -> tuple[str, int]:
     """
     Convert a 4-tuple IPv6 socket address to a 2-tuple (address, port) format.
 
@@ -498,6 +594,11 @@ def convert_ipv6_sockaddr(
     if isinstance(sockaddr, tuple) and len(sockaddr) == 4:
         host, port, flowinfo, scope_id = cast(Tuple[str, int, int, int], sockaddr)
         if scope_id:
+            # PyPy (as of v7.3.11) leaves the interface name in the result, so
+            # we discard it and only get the scope ID from the end
+            # (https://foss.heptapod.net/pypy/pypy/-/issues/3938)
+            host = host.split("%")[0]
+
             # Add scope_id to the address
             return f"{host}%{scope_id}", port
         else:
