@@ -1,29 +1,34 @@
+from __future__ import annotations
+
 import math
-from typing import Optional, Tuple, Type, TypeVar, overload
+from typing import Any, TypeVar, overload
 
 from ..streams.memory import (
-    MemoryObjectReceiveStream, MemoryObjectSendStream, MemoryObjectStreamState)
+    MemoryObjectReceiveStream,
+    MemoryObjectSendStream,
+    MemoryObjectStreamState,
+)
 
-T_Item = TypeVar('T_Item')
+T_Item = TypeVar("T_Item")
 
 
 @overload
 def create_memory_object_stream(
-    max_buffer_size: float, item_type: Type[T_Item]
-) -> Tuple[MemoryObjectSendStream[T_Item], MemoryObjectReceiveStream[T_Item]]:
+    max_buffer_size: float = ...,
+) -> tuple[MemoryObjectSendStream[Any], MemoryObjectReceiveStream[Any]]:
     ...
 
 
 @overload
 def create_memory_object_stream(
-    max_buffer_size: float = 0
-) -> Tuple[MemoryObjectSendStream, MemoryObjectReceiveStream]:
+    max_buffer_size: float = ..., item_type: type[T_Item] = ...
+) -> tuple[MemoryObjectSendStream[T_Item], MemoryObjectReceiveStream[T_Item]]:
     ...
 
 
 def create_memory_object_stream(
-    max_buffer_size: float = 0, item_type: Optional[Type[T_Item]] = None
-) -> Tuple[MemoryObjectSendStream, MemoryObjectReceiveStream]:
+    max_buffer_size: float = 0, item_type: type[T_Item] | None = None
+) -> tuple[MemoryObjectSendStream[Any], MemoryObjectReceiveStream[Any]]:
     """
     Create a memory object stream.
 
@@ -34,9 +39,9 @@ def create_memory_object_stream(
 
     """
     if max_buffer_size != math.inf and not isinstance(max_buffer_size, int):
-        raise ValueError('max_buffer_size must be either an integer or math.inf')
+        raise ValueError("max_buffer_size must be either an integer or math.inf")
     if max_buffer_size < 0:
-        raise ValueError('max_buffer_size cannot be negative')
+        raise ValueError("max_buffer_size cannot be negative")
 
     state: MemoryObjectStreamState = MemoryObjectStreamState(max_buffer_size)
     return MemoryObjectSendStream(state), MemoryObjectReceiveStream(state)
